package uk.ac.cam.ch.wwmm.opsin;

/**
 * The currently supported wordRules
 * All word rules mentioned here should have corresponding rules in StructureBuilder
 * @author dl387
 *
 */
enum WordRule{
	acetal,
	additionCompound,
	acidHalideOrPseudoHalide,
	acidReplacingFunctionalGroup,
	amineDiConjunctiveSuffix,
	anhydride,
	potentialAlcoholEster,
	carbonylDerivative,
	cyclicPeptide,
	divalentFunctionalGroup,
	ester,
	functionalClassEster,
	functionGroupAsGroup,
	glycol,
	glycolEther,
	monovalentFunctionalGroup,
	multiEster,
	oxide,
	polymer,
	simple,
	substituent
}