package uk.ac.cam.ch.wwmm.opsin;

/**
 * The supported wordTypes.
 * Adding further word types should not be taken lightly
 * @author dl387
 *
 */
enum WordType{
	full,
	polymer,
	substituent,
	functionalTerm,
}